jit.flush()
ac.setLogSilent(true)

local json = require('lib/json')
local genericUtils = require('src/generic_utils')
local dataFilename = ac.getTrackDataFilename('traffic.json')
local latestReplay = ac.getTrackDataFilename('trafficReplay.json')

local carlist = require('src.carpresets.metadata')
CarsList = require("src.carpresets." .. carlist["chosen"])

function ErrorPos(msg, pos)
  DebugShapes[msg] = pos
  error(msg)
end

-- Common module
package.add('src/common')

-- Editor module
package.add('src/editor')
local EditorMain = require('EditorMain')

local function loadData()
  -- return require('src/grid_generator')()
  return json.decode(io.load(dataFilename, '{}'))
end

local function loadRply()
  -- return require('src/grid_generator')()
  return json.decode(io.load(latestReplay, '{}'))
end

---@type EditorMain
-- local editor = EditorMain(require('src/grid_generator')())
local toolEditor = EditorMain(loadData())

-- Simulation module
package.add('src/simulation')
local TrafficSimulation = require('TrafficSimulation')
local TrafficConfig = require('TrafficConfig')
local TrafficDebugLayers = require('TrafficDebugLayers')

RANDSEED = 0
sim = ac.getSim()

SPAWN_POS = {
  indexes = {}
}

FIRSTSPAWN = true

simSettings = ac.storage{
  simulationSpeed = 1,
  clickToDelete = false,
  profileGC = false,
  debugLines = false,
  debugLayers = 'null',
  debugBehaviour = false,
  debugSpawnAround = false,
  carsNumber = 200,

  lodRange = 50,
  lodBrightness = 15,
  lodAngle = 80,
  enableReflects = false,

  lodDistance = 125,

  rearBrightness = 25,

  mostRecentRandSeed = 1,
  carPosSinceRst = vec3(),

  rstTO = vec3(),
  rstFROM = vec3(),

  enableLights = true,
}



local simTraffic, simBroken = nil, false
local simDebugLayers = TrafficDebugLayers(json.decode(simSettings.debugLayers))

local function syncTrafficConfig()
  TrafficConfig.debugBehaviour = simSettings.debugBehaviour
  TrafficConfig.debugSpawnAround = simSettings.debugSpawnAround
  TrafficConfig.driversCount = simSettings.carsNumber

  TrafficConfig.lodRange = simSettings.lodRange
  TrafficConfig.lodBrightness = simSettings.lodBrightness
  TrafficConfig.lodAngle = simSettings.lodAngle
  TrafficConfig.enableReflects = simSettings.enableReflects
end

local rcar = ac.getCar(0)
local function setOrLoadSeed()
  if not sim.isReplayActive then    
    local seed = math.round(ui.mousePos().x + 2000 * ui.mousePos().y, 0)

    simSettings.mostRecentRandSeed = seed
    math.randomseed(seed)
    --math.randomseed(123123123123)
    simSettings.carPosSinceRst = rcar.position

    print('generated random seed ', seed)
    ac.warn("car position is", simSettings.carPosSinceRst)
  else
    print('loading pseudo-random seed', simSettings.mostRecentRandSeed)

    local mag = rcar.position - simSettings.carPosSinceRst -- magnitude of the different points

    -- invert if they're negative
    if 0 > mag.x then mag.x = mag.x *-1 end
    if 0 > mag.y then mag.y = mag.y *-1 end
    if 0 > mag.z then mag.z = mag.z *-1 end

    -- assert that distance is lower than 0.001 on all axis
    local tolerance = 0.1
    if tolerance > mag.x and tolerance > mag.y then
      math.randomseed(simSettings.mostRecentRandSeed)
      --math.randomseed(123123123123)
      print("loaded seed ", simSettings.mostRecentRandSeed, "enjoy the replay")
      print("mag:",mag)
      
      return true
    else
      print("nmag:",mag)
    end
  end

end

local function trafficRefresh()

  local z = setOrLoadSeed()
  if simTraffic ~= nil then simTraffic:dispose() end
  simTraffic, simBroken = nil, false

  SPAWN_POS = {
    indexes = {}
  }

  _lastIndex = 0
  successfulIter = 0

  return z
end

trafficRefresh()

local text = 'hello world'
local comboSelectedIndex = 1

carsSlowedDown = 0
local activeTabObject = 'Main'

function script.windowMain(dt)
  
  ui.tabBar('##someTabBarID', function ()
    ui.tabItem('Main', function () mainUI(); activeTabObject = 'Main' end)
    ui.tabItem('Development', function () toolEditor:doUI(dt); activeTabObject = 'Development' end)
  end)
end

function mainUI()
  ui.header('Simulation speed')
  ui.pushFont(ui.Font.Small)

  local speed = refnumber(simSettings.simulationSpeed > 1 and 1 + (simSettings.simulationSpeed - 1) / 30 or simSettings.simulationSpeed)
  if ui.slider('##simulationSpeed', speed, 0, 2, string.format('Speed: %.1f', simSettings.simulationSpeed)) then
    simSettings.simulationSpeed = speed.value > 1 and 1 + (speed.value - 1) * 30 or speed.value
  end

  local lnewNumber, lchanged = ui.slider('##lodDistance', simSettings.lodDistance, 10, 2000, 'Render Distance: %.0fm', 2)
  if lchanged then
    simSettings.lodDistance = lnewNumber
  end

  if ui.button('Restart') then
    print('restart')
    trafficRefresh()
  end
  
  ui.sameLine()
  if ui.button('Reset speed') then
    simSettings.simulationSpeed = 1
  end

  ui.sameLine()
  if ui.button("Save Replay", ui.ButtonFlags.Disabled) then
    local fs = io.save(latestReplay, json.encode(SPAWN_POS))
    ui.toast(ui.Icons.File, "Saved JSON replay: "..tostring(fs))
  end

  
  ui.separator()
  ui.header("AI Headlights")

  local lodRangeNum, lodRangeChanged = ui.slider('##lodRange', simSettings.lodRange, 1, 250, string.format('AI Headlight Range: %.0f', simSettings.lodRange), 2)
  if lodRangeChanged then
    simSettings.lodRange = lodRangeNum
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  local lodBrightnessNum, lodBrightnessChanged = ui.slider('##lodBrightness', simSettings.lodBrightness, 1, 500, string.format('AI Headlight Brightness: %.0f', simSettings.lodBrightness), 2)
  if lodBrightnessChanged then
    simSettings.lodBrightness = lodBrightnessNum
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  local rearBrightness, rbrc = ui.slider('##rbrc', simSettings.lodBrightness, -25, 500, string.format('AI Tailight Brightness: %.0f', simSettings.rearBrightness), 2)
  if rbrc then
    simSettings.rearBrightness = rearBrightness
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  local lodAngleNum, lodAngleChanged = ui.slider('##lodAngle', simSettings.lodAngle, 1, 360, string.format('AI Headlight Angle: %.0f', simSettings.lodAngle), 2)
  if lodAngleChanged then
    simSettings.lodAngle = lodAngleNum
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  if ui.checkbox('Enable Lights in reflections', simSettings.enableReflects) then
    simSettings.enableReflects = not simSettings.enableReflects
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  if ui.checkbox('Enable Lights', simSettings.enableLights) then
    simSettings.enableLights = not simSettings.enableLights
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  if ui.button('Reset lights') then
    simSettings.lodRange = 24
    simSettings.lodBrightness = 12
    simSettings.lodAngle = 60
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  ui.separator()

  ui.header("Car List Picker")
  local comboIndex, comboChanged = ui.combo("##presetbox", comboSelectedIndex, ui.ComboFlags.None, carlist["carlists"])
  if comboChanged then

    comboSelectedIndex = comboIndex -- change index for combo to update...

    carlist["chosen"] = carlist["carlists"][comboIndex] -- and update the picked carslist

    -- update carslist
    CarsList = require("src.carpresets." .. carlist["chosen"])

    -- force reset
    trafficRefresh()
  end
  -- end of otter's edits

  ui.separator()

  if ui.checkbox('Profile GC', simSettings.profileGC) then
    simSettings.profileGC = not simSettings.profileGC
  end

  if ui.checkbox('Click to delete cars', simSettings.clickToDelete) then
    simSettings.clickToDelete = not simSettings.clickToDelete
  end

  if ui.checkbox('Debug cars behaviour', simSettings.debugBehaviour) then
    simSettings.debugBehaviour = not simSettings.debugBehaviour
    trafficRefresh()
  end

  if ui.checkbox('Spawn cars nearby only', simSettings.debugSpawnAround) then
    simSettings.debugSpawnAround = not simSettings.debugSpawnAround
    trafficRefresh()
  end

  local newNumber, changed = ui.slider('##carsNumber', simSettings.carsNumber, 1, 2000, 'Cars: %.0f', 2)
  if changed then
    simSettings.carsNumber = newNumber
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  if ui.button('Disable Cars') then
    simSettings.carsNumber = 0
    setTimeout(trafficRefresh, 0.01, 'refresh')
  end

  ui.separator()

  if ui.checkbox('Show debug lanes', simSettings.debugLines) then
    simSettings.debugLines = not simSettings.debugLines
  end

  local function debugLayer(layer)
    ui.offsetCursorX(layer.level * 8)
    if ui.checkbox(layer.name, layer.active) then
      layer.active = not layer.active
      simSettings.debugLayers = json.encode(simDebugLayers:serialize())
    end
    if layer.active then
      ui.pushID(layer.name)
      table.forEach(layer.children, debugLayer)
      ui.popID()
    end
  end

  if simSettings.debugLines then
    table.forEach(simDebugLayers.root.children, debugLayer)
  end

  if simBroken then
    local errorMsg = 'Sim has crashed'
    local solutionData = nil
    if ac.getLastError():match('Not allowed') then
      errorMsg = 'Can\'t run: scripting physics is not allowed. Add in “surfaces.ini”:'
      solutionData = '[SURFACE_0]\
WAV_PITCH=extended-0\
\
[_SCRIPTING_PHYSICS]\
ALLOW_TRACK_SCRIPTS=1\
ALLOW_DISPLAY_SCRIPTS=1\
ALLOW_NEW_MODE_SCRIPTS=1\
ALLOW_TOOLS=1'
    end

    ui.setCursor(vec2(28, 60))
    ui.pushStyleVar(ui.StyleVar.ChildRounding, 8)
    ui.pushStyleColor(ui.StyleColor.ChildBg, rgbm(0.5, 0, 0, 1))
    ui.childWindow('simErrorMsg', vec2(0, solutionData and 148 or 48), false, 0, function ()
      ui.offsetCursor(8)
      ui.text(errorMsg)
      if solutionData ~= nil then
        ui.offsetCursorX(8)
        ui.copyable(solutionData)
      end
      ui.offsetCursorX(8)
      if ui.button('Restart') then
        simSettings.simulationSpeed = 1
        trafficRefresh()
      end
      ui.sameLine()
      if ui.button('Resume') then
        simBroken = false
      end
    end)
    ui.popStyleColor()
    ui.popStyleVar()
  end

  ui.popFont()
end

local simulationTabName = 'Simulation'
--editor.tabs:insert(1, { name = simulationTabName, fn = script.windowMain })

-- Tool script
-- function script.asyncUpdate()
--   if not simSettings.profileGC then
--     collectgarbage()
--   end
-- end

local replaySeedLoaded = false

local replayDelay = 1 -- frames
local replayFrm = 0
function script.update(dt)

  if true then
    -- require('src/test/distance_between_cars')()
    -- require('src/test/perf_enum')()
    -- require('src/test/test_class')()
    -- require('src/test/test_dynamic_texture')()
    -- return
  end

  if sim.isReplayActive and replaySeedLoaded == false then
    --sim.repla

    if simTraffic ~= nil then simTraffic:dispose() end
    simTraffic, simBroken = nil, false

    local v = setOrLoadSeed()

    if v == true then
      ac.warn("this should have loaded the most recent seed")

      _lastIndex = 0
      SPAWN_POS = loadRply()
      if simTraffic ~= nil then simTraffic:dispose() end
      simTraffic, simBroken = nil, false

      print(SPAWN_POS)

      replaySeedLoaded = true
    end

    return
  elseif sim.isReplayActive == false and replaySeedLoaded == true then
    replaySeedLoaded = false 

  elseif sim.isReplayActive == true and replaySeedLoaded == true then
    if replayFrm ~= replayDelay then
      replayFrm = replayFrm + 1
      return
    end
  end

  -- take note of sim.dt
  if sim.isOnlineRace then
    -- Paused
    return
  end

  if sim.dt == 0 then
    return
  end

  if simSettings.simulationSpeed == 0 or simBroken then
    return
  end

  if not toolEditor:isEmpty() and simTraffic == nil then
    simBroken = true
    syncTrafficConfig()
    -- simTraffic = TrafficSimulation(editor:serializeData())
    simTraffic = TrafficSimulation(toolEditor:finalizeData())
    simBroken = false
  end

  if simSettings.profileGC then
    collectgarbage()
  end

  if sim.focusedCar ~= -1 then
    local car = ac.getCar(sim.focusedCar)
    if car ~= nil and not car.extrapolatedMovement then
      -- Hacky fix. Let’s hope we’ll be able to use extrapolated movement soon.
      dt = dt + (sim.time - ac.getCar(sim.focusedCar).timestamp) / 1e3
      dt = math.max(dt, 0.002)
    end
  end

  if simTraffic ~= nil and not simBroken then
    simBroken = true
    for _ = 1, math.ceil(simSettings.simulationSpeed) do
      simTraffic:update(math.saturateN(simSettings.simulationSpeed) * dt)
    end
    simBroken = false
  end

  if simSettings.profileGC then
    genericUtils.runGC()
  end

end

-- function update(dt)
--   editor:doUI(dt)
-- end

-- For easier debugging
DebugShapes = {}

function script.draw3D()
 
  if script.draw3DOverride then
    script.draw3DOverride()
  end

  if simTraffic then
    simTraffic:drawMain()
  end

  if activeTabObject == "Development" then
    render.setDepthMode(render.DepthMode.Off)
    pcall(function() toolEditor:draw3D() end)
    
  elseif simTraffic ~= nil and simSettings.debugLines then
    simDebugLayers:start()
    simTraffic:draw3D(simDebugLayers, simSettings.clickToDelete)

    table.forEach(DebugShapes, function (item, key)
      ac.debug(key, item)
      render.debugCross(item, 2, rgbm(3, 0, 0, 1))
      render.debugText(item, key, rgbm(3, 0, 0, 1))
    end)
  end


end

-- Updating simulation and saving on change
local refreshTimeout = nil
toolEditor.onChange:subscribe(function ()
  setTimeout(function ()
    io.save(dataFilename, json.encode(toolEditor:serializeData()))
  end, 3, 'save')

  clearTimeout(refreshTimeout)
  refreshTimeout = setTimeout(function ()
    trafficRefresh()
  end, 0.5)
end)

-- setInterval(function ()
--   trafficRefresh()
-- end, 0.3)

-- ac.debug('cfg', ac.getTrackConfig().number('CONDITION_1', 'DELAY', 0))

-- local v = ac.getTrackConfig('CONDITION_1', 'DELAY', 0)
-- ac.debug('cfg2', ac.getTrackConfig('CONDITION_1', 'DELAY', 0))
