local function loadColorsList(filename)
    local colors = {}
    for line in io.lines(filename) do
      table.insert(colors, rgbm.new(line))
    end
    return function () return table.random(colors) end
  end
  
  local colorRandom = loadColorsList('extension/config/data_oem_colors_modern.txt')
  local colorBland = loadColorsList('extension/config/data_oem_colors_vintage.txt')
  
  --- @class CarDefinitionLights
  --- @field headlights string[]
  --- @field rear string[]
  --- @field rearCombined string[]
  --- @field brakes string[]
  
  --- @class CarDefinitionDimensions
  --- @field front number
  --- @field rear number
  --- @field turningOffset number
  --- @field width number
  --- @field wheelRadius number
  --- @field fakeShadowX number
  --- @field fakeShadowZ number
  
  --- @class CarDefinitionPhysics
  --- @field mass number
  --- @field width number
  --- @field length number
  --- @field wheelsGripForce number
  --- @field suspensionTravel number
  --- @field suspensionForce number
  --- @field suspensionDamping number
  --- @field cog vec3|nil
  
  --- @class CarDefinition
  --- @field main string
  --- @field lod string
  --- @field collider string
  --- @field maxSpeed number
  --- @field dynamic number
  --- @field color fun(): rgbm
  --- @field lights CarDefinitionLights
  --- @field dimensions CarDefinitionDimensions
  --- @field physics CarDefinitionPhysics
  --- @field cache any
  
  local massMult = 1 -- set to 0.001 to have some fun with it, Driver ́’99 style (https://www.youtube.com/watch?v=DiN6OkklU1s)
  
  --- @type CarDefinition[]
  local cars = {
  
    {
      main = 'data/traffic_isuzu_npr/isuzu.kn5',
      lod = 'data/traffic_isuzu_npr/isuzu_lod_b.kn5',
      collider = 'data/traffic_isuzu_npr/collider.kn5',
      dynamic = 0.6,
      maxSpeed = 160,
      color = colorRandom,
      lights = {
        headlights = {'headlight'},
        rearCombined = {'Object001' , 'Object002' , 'lights_r2' , 'lights_r'} ,
        brakes = {'Object001' , 'Object002' , 'lights_r' , 'lights_r2'},
     LED = {'Object001'},
     LED_RGB = {24,4,0,},
      bodymaterial = 'body',
      --bodymesh = 'body',
      ksshader = 'txDetail', --'txDiffuse', --'txDetail',
      },
      dimensions = {
        front = 0.5,
        rear = 5,
        turningOffset = 2.3,
        width = 1.6,
        wheelRadius = 0.33,
        fakeShadowX = 0.8 + 0.35,
        fakeShadowZ = 2.25 + 0.15,
            hOffset = 0.25,
      },
      physics = {
        mass = 1500 * massMult,
        width = 1.5,
        length = 3,
        wheelsGripForce = 10,
        suspensionTravel = 0.1,
        suspensionForce = 30,
        suspensionDamping = 10
      },
  
      soundbank = "/data/audio/test.bank",
      cache = {}
    },
  }
  
  return cars
  