local function loadColorsList(filename)
  local colors = {}
  for line in io.lines(filename) do
    table.insert(colors, rgbm.new(line))
  end
  return function () return table.random(colors) end
end

local colorRandom = loadColorsList('extension/config/data_oem_colors_modern.txt')
local colorBland = loadColorsList('extension/config/data_oem_colors_vintage.txt')

--- @class CarDefinitionLights
--- @field headlights string[]
--- @field rear string[]
--- @field rearCombined string[]
--- @field brakes string[]

--- @class CarDefinitionDimensions
--- @field front number
--- @field rear number
--- @field turningOffset number
--- @field width number
--- @field wheelRadius number
--- @field fakeShadowX number
--- @field fakeShadowZ number

--- @class CarDefinitionPhysics
--- @field mass number
--- @field width number
--- @field length number
--- @field wheelsGripForce number
--- @field suspensionTravel number
--- @field suspensionForce number
--- @field suspensionDamping number
--- @field cog vec3|nil

--- @class CarDefinition
--- @field main string
--- @field lod string
--- @field collider string
--- @field maxSpeed number
--- @field dynamic number
--- @field color fun(): rgbm
--- @field lights CarDefinitionLights
--- @field dimensions CarDefinitionDimensions
--- @field physics CarDefinitionPhysics
--- @field cache any

local massMult = 1 -- set to 0.001 to have some fun with it, Driver ́’99 style (https://www.youtube.com/watch?v=DiN6OkklU1s)

--- @type CarDefinition[]
local cars = {

  {
    main = 'data/bmw_1m.kn5',
    lod = 'data/bmw_1m_lod.kn5',
    collider = 'data/bmw_1m_collider.kn5',
    dynamic = 0.8,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'front_light_1', 'front_light_2', 'front_light_1_SUB0' },
      rear = { 'rear_light_1' },
      brakes = { 'brake_light_1', 'brake_light_2' },
    },
    dimensions = { 
      front = 0.5, 
      rear = 4, 
      turningOffset = 2.3, 
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.0,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 7,
      suspensionTravel = 0.05,
      suspensionForce = 30,
      suspensionDamping = 15
    },
    cache = {}
  },


  {
    main = 'data/alfaromeo_giulietta.kn5',
    lod = 'data/alfaromeo_giulietta_lod.kn5',
    collider = 'data/alfaromeo_giulietta_collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'FRONT_LIGHT1', 'FRONT_LIGHT_LED', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'REAR_LIGHT_1', 'REAR_LIGHT_0' },
      brakes = {'REAR_LIGHT_2' , 'REAR_LIGHT_1', 'REAR_LIGHT_0'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.0,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/cadillac_escalade/cadillac_escalade.kn5',
    lod = 'data/cadillac_escalade/cadillac_escalade_lod.kn5',
    collider = 'data/cadillac_escalade/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r'},
      brakes = {'lights_r' , 'brake_top'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_chevy_imp/chevy_imp.kn5',
    lod = 'data/traffic_chevy_imp/chevy_imp_lod_c.kn5',
    collider = 'data/traffic_chevy_imp/collider.kn5',
    dynamic = 0.6,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights' },
      rearCombined = { 'lights_r' },
      brakes = {'brake_top' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.0,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/ks_porsche_macan_traffic/porsche_macan.kn5',
    lod = 'data/ks_porsche_macan_traffic/porsche_macan_lod_D.kn5',
    collider = 'data/ks_porsche_macan_traffic/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'LIGHT_4', 'LIGHT_1', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'REAR_LIGHT_2' },
      brakes = {'REAR_LIGHT_2' , 'REAR_LIGHT_4'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },


  {
    main = 'data/traffic_dodge_challenger_srt/dodge_challenger_srt.kn5',
    lod = 'data/traffic_dodge_challenger_srt/dodge_challenger_srt_lod.kn5',
    collider = 'data/traffic_dodge_challenger_srt/collider.kn5',
    dynamic = 0.9,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'lights_r' , 'REAR_LIGHT_2'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_toyota_prius/prius.kn5',
    lod = 'data/traffic_toyota_prius/prius_lod_d.kn5',
    collider = 'data/traffic_toyota_prius/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'lights_r' , 'REAR_LIGHT_2'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },


  {
    main = 'data/traffic_alfa_mito/mito.kn5',
    lod = 'data/traffic_alfa_mito/mito_lod_c.kn5',
    collider = 'data/traffic_alfa_mito/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights' },
      rearCombined = { 'lights_r' },
      brakes = {'brake_top' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.00,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_alfa_159/alfa_159.kn5',
    lod = 'data/traffic_alfa_159/alfa_159_lod_d.kn5',
    collider = 'data/traffic_alfa_159/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'REAR_LIGHT_2' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_range_rover_ss/rrss.kn5',
    lod = 'data/traffic_range_rover_ss/rrss_lod_d.kn5',
    collider = 'data/traffic_range_rover_ss/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'REAR_LIGHT_2' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_truck/truck_right_size_test.kn5',
    lod = 'data/traffic_truck/truck_right_size_test_lod.kn5',
    collider = 'data/traffic_truck/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = {'ID1527_002','ID1533_002','ID89_006' , 'ID89_004' , 'ID1587_002' , 'ID1607_002' , 'ID63' , 'ID89' , 'ID89_002' , 'ID89_008' , 'ID1717_002' , 'ID2435_003'},
      rearCombined = {'ID3842','ID3848','ID3998','ID4006' , 'ID3932' , 'ID3938-1'} ,
      brakes = {'ID3836','ID3854','ID3842','ID3848','ID3998','ID4006'} ,
    },
    dimensions = { 
      front = 2.0, 
      rear = 4.5, 
      turningOffset = 2.8, 
      width = 2.2,
      wheelRadius = 0.48,
      fakeShadowX = 1.2 + 0.35,
      fakeShadowZ = 6.2 + 0.35,
    },
    physics = {
      mass = 2500 * massMult,
      width = 2.6,
      length = 12.3,
      wheelsGripForce = 7,
      suspensionTravel = 0.05,
      suspensionForce = 30,
      suspensionDamping = 15
    },
    cache = {}
  },


  {
    main = 'data/traffic_volvo_v70/volvo_v70.kn5',
    lod = 'data/traffic_volvo_v70/volvo_v70_lod_d.kn5',
    collider = 'data/traffic_volvo_v70/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'brake_top' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },


  {
    main = 'data/audi_a4/audi_a4.kn5',
    lod = 'data/audi_a4/audi_a4_lod_d.kn5',
    collider = 'data/audi_a4_collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights' },
      rearCombined = { 'lights_r' },
      brakes = {'REAR_LIGHT_2' , 'lights_r' },
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },
}

return cars
