local function loadColorsList(filename)
  local colors = {}
  for line in io.lines(filename) do
    table.insert(colors, rgbm.new(line))
  end
  return function () return table.random(colors) end
end

local colorRandom = loadColorsList('extension/config/data_oem_colors_modern.txt')
local colorBland = loadColorsList('extension/config/data_oem_colors_vintage.txt')

--- @class CarDefinitionLights
--- @field headlights string[]
--- @field rear string[]
--- @field rearCombined string[]
--- @field brakes string[]

--- @class CarDefinitionDimensions
--- @field front number
--- @field rear number
--- @field turningOffset number
--- @field width number
--- @field wheelRadius number
--- @field fakeShadowX number
--- @field fakeShadowZ number

--- @class CarDefinitionPhysics
--- @field mass number
--- @field width number
--- @field length number
--- @field wheelsGripForce number
--- @field suspensionTravel number
--- @field suspensionForce number
--- @field suspensionDamping number
--- @field cog vec3|nil

--- @class CarDefinition
--- @field main string
--- @field lod string
--- @field collider string
--- @field maxSpeed number
--- @field dynamic number
--- @field color fun(): rgbm
--- @field lights CarDefinitionLights
--- @field dimensions CarDefinitionDimensions
--- @field physics CarDefinitionPhysics
--- @field cache any

local massMult = 1 -- set to 0.001 to have some fun with it, Driver ́’99 style (https://www.youtube.com/watch?v=DiN6OkklU1s)

--- @type CarDefinition[]
local cars = {




  {
    main = 'data/bmw_e30_w_traffic/bmw_e30_w.kn5',
    lod = 'data/bmw_e30_w_traffic/bmw_e30_w_lod.kn5',
    collider = 'data/bmw_e30_w_traffic/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'LIGHT_FRONT', 'LIGHT_FRONT2', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'LIGHT_TAIL'},
      brakes = {'LIGHT_TAIL' , 'LIGHT_BRAKE'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.0,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_isuzu_npr/isuzu.kn5',
    lod = 'data/traffic_isuzu_npr/isuzu_lod_b.kn5',
    collider = 'data/traffic_isuzu_npr/collider.kn5',
    dynamic = 0.6,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = {'headlight'},
      rearCombined = {'Object001' , 'Object002' , 'lights_r2' , 'lights_r'} ,
      brakes = {'Object001' , 'Object002' , 'lights_r' , 'lights_r2'},
	 LED = {'Object001'},
	 LED_RGB = {24,4,0,},
	  bodymaterial = 'body',
	  --bodymesh = 'body',
	  ksshader = 'txDetail', --'txDiffuse', --'txDetail',
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.25,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },


  {
    main = 'data/traffic_isuzu_haulerb/isuzu.kn5',
    lod = 'data/traffic_isuzu_haulerb/isuzu_lod.kn5',
    collider = 'data/traffic_isuzu_haulerb/collider.kn5',
    dynamic = 0.6,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = {'headlight', 'r4'},
      rearCombined = {'l4' ,'r1', 'sidelights','lights_r','lights_r2' } ,
      brakes = {'l4' ,'r1', 'sidelights' , 'lights_r','lights_r2' },
	 LED = {'r1'},
	 LED_RGB = {24,4,0,},
	  bodymaterial = 'body',
	  --bodymesh = 'body',
	  ksshader = 'txDetail', --'txDiffuse', --'txDetail',
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.25,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_isuzu_tanker/isuzu.kn5',
    lod = 'data/traffic_isuzu_tanker/isuzu_lod_b.kn5',
    collider = 'data/traffic_isuzu_tanker/collider.kn5',
    dynamic = 0.6,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = {'headlight'},
      rearCombined = {'sidelights2','lights_r2', 'lights_r' } ,
      brakes = {'sidelights2' , 'lights_r2', 'lights_r' },
      LED = {'r1'},
	 LED = {'sidelights2'},
	 LED_RGB = {24,4,0,},
	  bodymaterial = 'body',
	  --bodymesh = 'body',
	  ksshader = 'txDetail', --'txDiffuse', --'txDetail',
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.25,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },



  {
    main = 'data/traffic_toyota_camry/camry.kn5',
    lod = 'data/traffic_toyota_camry/camry_lod_d.kn5',
    collider = 'data/traffic_toyota_camry/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = { 'headlights', --[[ 'polymsh_detached1_SUB4' ]] },
      rearCombined = { 'lights_r' },
      brakes = {'lights_r' , 'REAR_LIGHT_2'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_mercedes_sprinter/mercedes_sprinter_rhd.kn5',
    lod = 'data/traffic_mercedes_sprinter/mercedes_sprinter_rhd_lod_b.kn5',
    collider = 'data/traffic_mercedes_sprinter/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'ID735_003','ID357_003' },
      rearCombined = { 'ID681_003','ID643_003' },
      brakes = {'ID681_003','ID643_003' , 'ID383_003','ID363_003'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.0,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

  {
    main = 'data/traffic_truck/truck_right_size_test.kn5',
    lod = 'data/traffic_truck/truck_right_size_test_lod.kn5',
    collider = 'data/traffic_truck/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorBland,
    lights = {
      headlights = {'ID1527_002','ID1533_002','ID89_006' , 'ID89_004' , 'ID1587_002' , 'ID1607_002' , 'ID63' , 'ID89' , 'ID89_002' , 'ID89_008' , 'ID1717_002' , 'ID2435_003'},
      rearCombined = {'ID3842','ID3848','ID3998','ID4006' , 'ID3932' , 'ID3938-1'} ,
      brakes = {'ID3836','ID3854','ID3842','ID3848','ID3998','ID4006'} ,
    },
    dimensions = { 
      front = 2.0, 
      rear = 4.5, 
      turningOffset = 2.8, 
      width = 2.2,
      wheelRadius = 0.48,
      fakeShadowX = 1.2 + 0.35,
      fakeShadowZ = 6.2 + 0.35,
    },
    physics = {
      mass = 2500 * massMult,
      width = 2.6,
      length = 12.3,
      wheelsGripForce = 7,
      suspensionTravel = 0.05,
      suspensionForce = 30,
      suspensionDamping = 15
    },
    cache = {}
  },


  {
    main = 'data/traffic_toyota_rav4/rav4.kn5',
    lod = 'data/traffic_toyota_rav4/rav4_lod_b.kn5',
    collider = 'data/traffic_toyota_rav4/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = { 'headlights' },
      rearCombined = { 'lights_r' },
      brakes = {'brake_top' , 'lights_r'},
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.33,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
          hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.1,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },
  

  {
    main = 'data/traffic_nissan_leaf/traffic_nissan_leaf.kn5',
    lod = 'data/traffic_nissan_leaf/traffic_nissan_leaf_lod_d.kn5',
    collider = 'data/traffic_nissan_leaf/collider.kn5',
    dynamic = 0.4,
    maxSpeed = 160,
    color = colorRandom,
    lights = {
      headlights = {'headlights'},
      rearCombined = {'lights_r'} ,
      brakes = {'brake_top' , 'lights_r'},
	 --LED = {'top_light','taxiataxi_lights',},
	 --LED_RGB = {8,6,2,},
	  bodymaterial = 'body',
	  --bodymesh = 'body',
	  ksshader = 'txDetail', --'txDiffuse', --'txDetail',
    },
    dimensions = {
      front = 0.5,
      rear = 4,
      turningOffset = 2.3,
      width = 1.6,
      wheelRadius = 0.5,
      fakeShadowX = 0.8 + 0.35,
      fakeShadowZ = 2.25 + 0.15,
      hOffset = 0.15,
    },
    physics = {
      mass = 1500 * massMult,
      width = 1.5,
      length = 3,
      wheelsGripForce = 10,
      suspensionTravel = 0.3,
      suspensionForce = 30,
      suspensionDamping = 10
    },
    cache = {}
  },

}

return cars
